# -*- coding: utf-8 -*-
"""
Created on Fri Mar 23 11:33:24 2018

@author: stuart
"""

#single file processing tools for nanowire-plugin

import os
from minio import Minio
import time
import traceback
import logging
import json
from os import environ
import inspect

import requests

#from ssl import PROTOCOL_TLSv1_2



import sys
#import hashlib
import datetime

#import urllib.request

from nanowire_plugin import send

#from minio.error import AccessDenied


###############################################################################
### Here the tools that could be exported to single file plugin can be sent ###
###############################################################################
#!!!!!!!!!!!!!!!!!!!!!!!!

#set up the logger globally
logger = logging.getLogger("nanowire-plugin")


#logging.basicConfig(level=10, stream=sys.stdout)

class Worker(object):
    def __init__(self, function, debug_mode, monitor_url, minio_client):
        
        self.function = function
        self.minio_client = minio_client
        self.monitor_url = monitor_url
        self.debug_mode = debug_mode

        logging.debug("ESTABLISHED WORKER")

    def run(self):
        while True:
            
            message = requests.get(os.environ['CONTROLLER_BASE_URI'] + '/v1/tasks/?pluginId=' + os.environ['PLUGIN_ID'] + '&pluginInstance=' + os.environ['POD_NAME'])
            
            #print("-------------------")
            #print(message.status_code)
            #print(dir(message))
            #print(message.text)
            #print("-------------------")
            code = message.status_code
            
            if code == 200:
                
                payload = json.loads(message.text)
                
                meta = payload['metadata']
                jsonld = payload['jsonld']
                try:
                    url = meta['task']['metadata']['cacheURL']
                except:
                    url = None
                    
                if meta['job']['workflow']['type'] == 'GROUP':
                    logger.warning("SINGLE FILE PLUGIN TOOL WAS SENT A GROUP JOB")
                    
                else:
                    #try to run our function
                    try:
                        #result = self.function(metadata, jsonld, url)
                        result = run_function(self.function, meta, jsonld, url)
                        
                    except Exception as exp:
                        if self.debug_mode > 0:
                            result = str(traceback.format_exc())
                            logger.info("THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s"%result)
                        else:
                            result = str(exp)
                            logger.info("THERE WAS A PROBLEM RUNNING THE MAIN FUNCTION: %s"%result)
                            

                    job_stats = send(meta, result, self.minio_client, self.debug_mode)
                    
                    
                    if self.debug_mode >= 1:
                        logger.info(job_stats)
                            
                    logger.info("FINISHED RUNNING USER CODE AT %s"%str(datetime.datetime.now()))
                    logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
                    
                            
                
            elif code == 404:
                time.sleep(1)


def bind(function, debug_mode=0):
    """binds a function to the input message queue"""
    
    #set up the logging
    logger.setLevel(logging.DEBUG)
    
    if debug_mode > 0:
        #write to screen to ensure logging is working ok
        #print "Initialising nanowire lib, this is a print"
        logger.info("Running on %s"%sys.platform)
    
    
    #set up the minio client. Do this before the AMPQ stuff
    
    if 'MINIO_REGION' in environ.keys():
        minio_client = Minio(
            environ["MINIO_HOST"] + ":" + environ["MINIO_PORT"],
            access_key=environ["MINIO_ACCESS"],
            secret_key=environ["MINIO_SECRET"],
            region = environ['MINIO_REGION'],
            secure=True if environ["MINIO_SCHEME"] == "https" else False)
    else:
        minio_client = Minio(
            environ["MINIO_HOST"] + ":" + environ["MINIO_PORT"],
            access_key=environ["MINIO_ACCESS"],
            secret_key=environ["MINIO_SECRET"],
            region=None,
            secure=True if environ["MINIO_SCHEME"] == "https" else False)
        
    '''
    #use a boto client to try and avoid some serious bugs with minio
    session = boto3.session.Session()
    
    minio_client = session.client('s3', region_name='ams3', endpoint_url=environ['MINIO_HOST']+":"+environ['MINIO_PORT'],
                                  aws_access_key_id=environ['MINIO_ACCESS'],
                                    aws_secret_access_key=environ['MINIO_SECRET'],
                                    secure=True if environ['MINIO_SCHEME']=='https' else False)    
    '''
    #minio_client.set_app_info(name, '1.0.0')

    monitor_url = environ["MONITOR_URL"]

    logger.info("initialised nanowire lib", extra={
        "monitor_url": monitor_url,
        "minio": environ["MINIO_HOST"],
        "rabbit": environ["AMQP_HOST"]
    })
    
    logger.info("monitor_url: %s"%monitor_url)
    logger.info("minio: %s"%environ["MINIO_HOST"])
    logger.info("rabbit: %s"%environ["AMQP_HOST"])


    #this is only commented out since I'm trying to find the source of these terrible errors
    
    worker = Worker(function, debug_mode, monitor_url, minio_client)
    worker.run()
    logger.warning("PAST THE RUN FUNCTION, SOMETHING HAS GONE VERY WRONG")
        
        
        

def validate_single_file_function(function):
    
    if sys.version_info.major == 3:
        
        arguments = list(inspect.signature(function).parameters)
      
    elif sys.version_info.major == 2:
        
        arguments = inspect.getargspec(function)[0]
        
    allowed = ['self', 'jsonld', 'metadata', 'url']
    
    arg_dict = set()
    
    for arg in arguments:
        
        if arg not in arg_dict:
            arg_dict.add(arg)
        else:
            raise Exception("ARGUMENTS MAY NOT BE REPEATED")
            
        if arg not in allowed:
            raise Exception("FUNCTION MAY ONLY USE ALLOWED ARGUMENTS, ALLOWED ARGUMENTS ARE: jsonld, metadata, url, YOU HAVE USED THE ARGUMENT %s"%arg)
    
    if 'jsonld' not in arguments:
        
        raise Exception("FUNCTION MUST TAKE jsonld AS AN ARGUMENT")
        

def run_function(function, metadata, jsonld, url):


    if sys.version_info.major == 3:
      
        arguments = inspect.signature(function).parameters
        
    elif sys.version_info.major == 2:
        
        arguments = inspect.getargspec(function)[0]
    
    arguments = inspect.getargspec(function)[0]
    
    #3 argument variations
    if arguments == ['metadata', 'jsonld', 'url'] or arguments == ['self', 'metadata', 'jsonld', 'url']:
        
        result = function(metadata, jsonld, url)
        
        return result
        
    elif arguments == ['metadata', 'url', 'jsonld'] or arguments == ['self', 'metadata', 'url', 'jsonld']:
        
        result = function(metadata, url, jsonld)
        
        return result
        
    elif arguments == ['jsonld', 'metadata', 'url'] or arguments == ['self', 'jsonld', 'metadata', 'url']:
        
        result = function(jsonld, metadata, url)
        
        return result
        
    elif arguments == ['jsonld', 'url', 'metadata'] or arguments == ['self', 'jsonld', 'url', 'metadata']:
        
        result = function(jsonld, url, metadata)
        
        return result
        
    elif arguments == ['url', 'metadata', 'jsonld'] or arguments == ['self', 'url', 'metadata', 'jsonld']:
        
        result = function(url, metadata, jsonld)
        
        return result
        
    elif arguments == ['url', 'jsonld', 'metadata'] or arguments == ['self', 'url', 'jsonld', 'metadata']:
        
        result = function(url, jsonld, metadata)
        
        return result
        
        
    #2 argument variations
    elif arguments == ['url', 'jsonld'] or arguments == ['self', 'url', 'jsonld']:
        
        result = function(url, jsonld)
        
        return result

    elif arguments == ['jsonld', 'url'] or arguments == ['self', 'jsonld', 'url']:
        
        result = function(jsonld, url)
        
        return result

    elif arguments == ['metadata', 'jsonld'] or arguments == ['self', 'metadata', 'jsonld']:
        
        result = function(metadata, jsonld)
        
        return result
        
    elif arguments == ['jsonld', 'metadata'] or arguments == ['self', 'jsonld', 'metadata']:
        
        result = function(jsonld, metadata)
        
        return result
        
    #1 argument variations
    elif arguments == ['jsonld'] or arguments == ['self', 'jsonld']:
        
        result = function(jsonld)
        
        return result


    else:
        
        raise Exception("FUNCTION MUST ACCEPT VALID ARGUMENTS, CURRENT ARGUMENTS ARE %s"%str(arguments))



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 25 11:30:46 2017

@author: stuart
"""

#!/usr/bin/env python3
"""
Provides a `bind` function to plugins so they can simply bind a function to a queue.
"""

import traceback
import logging
import json
from os import environ
from os.path import join

import uuid

import datetime

#from ssl import PROTOCOL_TLSv1_2

#from minio import Minio
import os
import time
import sys

#import the relavant version of urllib depending on the version of python we are
if sys.version_info.major == 3:
    import urllib
elif sys.version_info.major == 2:
    import urllib
    import urllib2
else:
    import urllib
    
    
import requests

#import hashlib

#set up the logger globally
logger = logging.getLogger("nanowire-plugin")



##################################################################
### These tools are used by both group and single file plugins ###
##################################################################

def validate_payload(payload):
    """ensures payload includes the required metadata and this plugin is in there"""

    if not isinstance(payload, dict):
        raise Exception("payload is a %s, not a dictionary"%type(payload))

    if "nmo" not in payload:
        raise Exception("No nmo in payload")

    if "job" not in payload["nmo"]:
        raise Exception("No job in nmo \nnmo is %s"%payload["nmo"])

    if "task" not in payload["nmo"]:
        raise Exception("No task in nmo \nnmo is %s"%payload["nmo"])
    
    try:
        isGroup = payload['nmo']['source']['misc']['isGroup']    
    except:
        isGroup = False
        
        
    if "jsonld" not in payload and not isGroup:
        raise Exception("No jsonld in payload \nPayload is:- %s"%payload)
        



#Send the data back to the controler
def send(metadata, output, minio_client, debug_mode):
    """unwraps a message and calls the user function"""   
    
    #log some info about what the send function has been given
    logger.info("LETTING MONITOR KNOW PROCESSING HAS BEEN DONE")
        
        
    if isinstance(output, str):
        err = output
        output = {}
    elif isinstance(output, dict):
        err = None
    elif output == None:
        output = {}
        err = "NO OUTPUT WAS RETURNED"
    
    #send the info from this plugin to the next one in the pipeline
    send_result(output, metadata, err)
    
   

def send_result(output, metadata, error):
    
    logger.info("RUNNING SEND TO NEXT PLUGIN")

    #check if we're dealing with a group job. If it is a group job
    #any additional metadata created will be stored in what is erroneously
    #labeled as jsonld to make life easier
    if metadata['job']['workflow']['type'] == 'GROUP':
        
        #Group job went without a hitch
        if error == None:
            
            logger.debug("NO ERRORS DETECTED RETURNING FINE JSONLD")
            
            send_json = {
                  "pluginInstance": os.environ["POD_NAME"],
                  "status": "success",
                  "additionalMetadata": output
                }
        #this is the instance of we had an error when calculating the group
        #job
        else:
            
            logger.debug("ERROR WAS DETECTED")
            logger.debug(error)
            
            send_json = {
                  "pluginInstance": os.environ["POD_NAME"],
                  "status": "failure",
                  'error':error,
                  "additionalMetadata": output
                }
                
            #logger.debug("~~~~~~~~~~~~~~~~~~~~~~~~~")
            #logger.debug(json.dumps(send_json))
        
    #if we don't have a group job we must have a single file job.
    else:
        
        if error == None:
            
            send_json = {'pluginInstance':os.environ['POD_NAME'],
                         'status':'success',
                         'jsonld':output}
                         
        else:
            #logger.warning(error)
            #logger.warning("++++++++++++++++++++++")
            send_json = {'pluginInstance':os.environ['POD_NAME'],
             'status':'failure',
             'error':error,
             'jsonld':output}
             
    #This bit sends the result to the controller. It is the most likely
    #source of trouble
    r = requests.put(os.environ['CONTROLLER_BASE_URI'] + '/v1/tasks/' + metadata['task']['_id'],
                     data=json.dumps(send_json), headers={'content-type':'application/json'})
    
    if r.status_code != 200:
        logger.warning(str(dir(r)))
        logger.warning(str(r.reason))
        logger.warning(str(r.text))
        logger.warning("-----------")
        logger.warning(str(output))
        logger.warning(str(type(output)))
        logger.warning("---------")
        logger.warning("PUT REQUEST STATUS IS %s SOMETHING HAS GONE VERY WRONG"%str(r.status_code))

  
    
def grab_file(url):
    
    if not os.path.exists('/tmp'):
        os.mkdir('/tmp')
    try:
        filename = '/tmp/document'
        urllib.request.urlretrieve(url, filename)
        return filename
    except:
        
        return "COULD NOT GRAB DOCUMENT, URL MAY HAVE EXPIRED"
        


#EXPEREMENTAL SECTION
def create_task(nmo, new_jsonld):
    
    job_id = nmo["job"]["job_id"],
    
    monitor_url = environ["MONITOR_URL"]
    
    task_dict = {}
    task_dict['_id'] = str(uuid.uuid4())
    task_dict["name"] = "Extracted data file"
    task_dict["type"] = "TestAction"
    task_dict["jobId"] = job_id
    
    upload_dict = {}
    upload_dict["type"] = "Example"
    upload_dict["meta"] = new_jsonld
    upload_dict["credentials"] = {"accessToken":environ["AMQP_USER"],
                                "accessTokenSecret":environ["AMQP_PASS"],
                                "consumerKey":environ["AMQP_USER"],
                                "consumerSecret":environ["AMQP_PASS"]}
    payload=json.dumps({
        "task":task_dict,
        "upload":upload_dict})
        
    #if we're working with python3
    if sys.version_info.major == 3:
        
        #if debug_status >= 2:
        #    logger.info("Running in python 3")
        
        request_url = monitor_url + "/v1/tasks"
        #urllib.parse.urljoin(monitor_url,"/v4/tasks/%s/positions"%task_id)
        
        req = urllib.request.Request(request_url,
            payload.encode(),
            headers={
                "Content-Type": "application/json"
            })
            
        result = urllib.request.urlopen(req)
        
        #logger.info("Task creation result was")
        #logger.info(str(result))

    
    #if we're working with python2
    elif sys.version_info.major == 2:
        
        #if debug_status >= 2:
        #    logger.info("Running in python 2")
        
        #there's no urljoin command in python2
        request_url = monitor_url + "/v1/tasks"       
        
        req = urllib2.Request(request_url,
            payload.encode(),
            headers={
                "Content-Type": "application/json"
            })
            
        result = urllib2.urlopen(req)
        
        logger.info("Task creation result was")
        logger.info(str(result))
    
    #if we're not in python2 or python3
    else:
        
        logger.warning("Running in an unknown version of python:- %s"%str(sys.version_info))